
! random forest convection emulator

module rf_convection_mod


!----------------------------------------------------------------------
use            fms_mod, only:  file_exist, error_mesg, open_file,  &
                               check_nml_error, mpp_pe, FATAL,  &
                               close_file
use simple_sat_vapor_pres_mod, only:  escomp, descomp
use      constants_mod, only:  HLv,HLs,Cp_air,Grav,rdgas,rvgas, &
                               kappa
use netcdf

implicit none
private


!---------------------------------------------------------------------
!  ---- public interfaces ----

   public  rf_convection, rf_convection_init, check

!-----------------------------------------------------------------------
!   ---- version number ----

 character(len=128) :: version = '$Id: rf_convection.f90, Paul OGorman, 2017 $'
 character(len=128) :: tag = '$Name: fez $'

!-----------------------------------------------------------------------
!   ---- local/private data ----

    logical :: do_init=.true.
    integer :: n_features ! number of features
    integer :: max_n_nodes ! maximum number of nodes across trees
    integer :: n_trees  ! number of trees
    integer :: n_outputs  ! number of outputs
    integer :: n_lev_rf  ! number of vertical levels the random forest uses

    integer(4), allocatable, dimension(:,:)  :: children_left
    integer(4), allocatable, dimension(:,:)  :: children_right
    integer(4), allocatable, dimension(:,:)  :: split_feature
    real(4), allocatable, dimension(:,:)     :: threshold
    real(4), allocatable, dimension(:,:,:)   :: values_predicted
    integer(4), allocatable, dimension(:)    :: n_nodes
    real(4), allocatable, dimension(:)       :: xscale_mean
    real(4), allocatable, dimension(:)       :: xscale_stnd


! do_init           has the module been initialized

! children_left     left child of node (n_trees, max_n_nodes)
! children_right    right child of node
! split_feature     feature used for splitting the node
! threshold         threshold value at the node
! values_predicted  prediction for a given node (n_trees, max_n_nodes, n_lev)
! xscale_mean       offset using in scaling of input variables
! xscale_stnd       factor using in scaling of input variables
! n_nodes           number of nodes in a given tree (n_trees)


! Namelist:
! rf_filename       data for random forest
! no_ps             set scaled surface pressure to a constant (zero)
! do_rh             use relative humidity instead of specific humidity


!-----------------------------------------------------------------------
!   ---- namelist ----

character(len=256) :: rf_filename = 'INPUT/rf_trees_v1.nc'
logical            :: no_ps = .false.
logical            :: do_rh = .false.
logical            :: x_scaler = .false. ! not needed with RF

namelist /rf_convection_nml/ rf_filename, no_ps, x_scaler, do_rh

contains

!#######################################################################

   subroutine rf_convection_init

!-----------------------------------------------------------------------
!
!        initialization for rf convection
!
!-----------------------------------------------------------------------
integer  unit,io,ierr

! This will be the netCDF ID for the file and data variable.
integer :: ncid
integer :: n_nodes_varid, children_left_varid, children_right_varid
integer :: trees_dimid, features_dimid, outputs_dimid, nodes_dimid
integer :: split_feature_varid, threshold_varid, values_predicted_varid
integer :: xscale_mean_varid, xscale_stnd_varid

  
!----------- read namelist ---------------------------------------------

      if (file_exist('input.nml')) then
         unit = open_file (file='input.nml', action='read')
         ierr=1; do while (ierr /= 0)
            read  (unit, nml=rf_convection_nml, iostat=io, end=10)
            ierr = check_nml_error (io,'rf_convection_nml')
         enddo
  10     call close_file (unit)
      endif

!---------- output namelist --------------------------------------------

      unit = open_file (file='logfile.out', action='append')
      if ( mpp_pe() == 0 ) then
           write (unit,'(/,80("="),/(a))') trim(version), trim(tag)
           write (unit,nml=rf_convection_nml)
      endif
      call close_file (unit)

!-------------allocate arrays and read data-------------------------


      if ( mpp_pe() == 0 ) then
       write(*,*) rf_filename
      endif

! Open the file. NF90_NOWRITE tells netCDF we want read-only access
! Get the varid or dimid for each variable or dimension based on its name.

      call check( nf90_open(     trim(rf_filename),NF90_NOWRITE,ncid ))

      call check( nf90_inq_dimid(ncid, 'dim_trees', trees_dimid))
      call check( nf90_inquire_dimension(ncid, trees_dimid, len=n_trees))

      call check( nf90_inq_dimid(ncid, 'dim_features', features_dimid))
      call check( nf90_inquire_dimension(ncid, features_dimid, len=n_features))

      call check( nf90_inq_dimid(ncid, 'dim_outputs', outputs_dimid))
      call check( nf90_inquire_dimension(ncid, outputs_dimid, len=n_outputs))

      call check( nf90_inq_dimid(ncid, 'dim_nodes', nodes_dimid))
      call check( nf90_inquire_dimension(ncid, nodes_dimid, len=max_n_nodes))

      allocate(n_nodes(n_trees))

      call check( nf90_inq_varid(ncid, "n_nodes", n_nodes_varid))
      call check( nf90_get_var(ncid, n_nodes_varid, n_nodes))

! basic checks of the data structure
      if(n_outputs /= n_features-1) call error_mesg ('rf_convection',  &
                        'n_outputs not matching n_features-1.', FATAL)

      if(max_n_nodes /= maxval(n_nodes)) call error_mesg ('rf_convection',  &
                        'max_n_nodes not equal to maximum of n_nodes', FATAL)

      n_lev_rf = (n_features-1)/2

      allocate(children_left(n_trees, max_n_nodes))
      allocate(children_right(n_trees, max_n_nodes))
      allocate(split_feature(n_trees, max_n_nodes))
      allocate(threshold(n_trees, max_n_nodes))
      allocate(values_predicted(n_trees, max_n_nodes, 2*n_lev_rf))

      if (x_scaler) then
       allocate(xscale_mean(n_features))
       allocate(xscale_stnd(n_features))
      endif

      call check( nf90_inq_varid(ncid, "children_left", children_left_varid))
      call check( nf90_get_var(ncid, children_left_varid, children_left))
      
      call check( nf90_inq_varid(ncid, "children_right", children_right_varid))
      call check( nf90_get_var(ncid, children_right_varid, children_right))

      call check( nf90_inq_varid(ncid, "split_feature", split_feature_varid))
      call check( nf90_get_var(ncid, split_feature_varid, split_feature))

      call check( nf90_inq_varid(ncid, "threshold", threshold_varid))
      call check( nf90_get_var(ncid, threshold_varid, threshold))

      call check( nf90_inq_varid(ncid, "values_predicted", values_predicted_varid))
      call check( nf90_get_var(ncid, values_predicted_varid, values_predicted))
    
      if (x_scaler) then
       call check( nf90_inq_varid(ncid,"xscale_mean",     xscale_mean_varid))
       call check( nf90_get_var(  ncid, xscale_mean_varid,xscale_mean      ))

       call check( nf90_inq_varid(ncid,"xscale_stnd",     xscale_stnd_varid))
       call check( nf90_get_var(  ncid, xscale_stnd_varid,xscale_stnd      ))
      endif
      
    
! Close the file
      call check( nf90_close(ncid))

      write(*, *) 'Finished reading regression file.'


! change from python to f90 convection for array indexing
      children_left = children_left + 1
      children_right = children_right + 1
      split_feature = split_feature + 1

      do_init=.false.
   end subroutine rf_convection_init


!#######################################################################

   subroutine rf_convection (tin, qin, pfull, phalf,  &
                             rain, tdel, qdel)

!-----------------------------------------------------------------------
!
!       Random-forest emulation of convection scheme
!
!-----------------------------------------------------------------------
!
!   input:  tin               temperature at full model levels
!           qin               specific humidity of water vapor at full
!                               model levels
!           pfull             pressure at full model levels
!           phalf             pressure at half (interface) model levels
!
!  note outputs are tendencies rather than adjustments
!  output:  rain     precipitation rate
!           tdel     temperature tendency at full model levels
!           qdel     specific humidity tendency (of water vapor) at
!                      full model levels
!
!-----------------------------------------------------------------------


!--------------------- interface arguments -----------------------------

   real   , intent(in) , dimension(:,:,:) :: tin, qin, pfull, phalf
   real   , intent(out), dimension(:,:)   :: rain
   real   , intent(out), dimension(:,:,:) :: tdel, qdel

!-----------------------------------------------------------------------
!---------------------- local data -------------------------------------


   real(4), dimension(n_features)             :: features
   real(4), dimension(2*n_lev_rf)             :: outputs

   real, dimension(size(tin,1),size(tin,2),size(tin,3)) :: esat, e, rh
   real:: e0,T0,eps

   integer  i, j, k, ix, jx, kx, ktop, krf, inode, itree, next_node


   if (do_init) call error_mesg ('rf_convection',  &
                        'rf_convection_init has not been called.', FATAL)

   ix=size(tin,1)
   jx=size(tin,2)
   kx=size(tin,3)
   krf = kx - n_lev_rf + 1  ! Index value to grab data for RF from full profile

   if (do_rh) then 
    call escomp(tin,esat)
    eps = rdgas/rvgas;
    e = pfull*qin/(eps+(1.0-eps)*qin)
    rh = e/esat
   endif


   do i=1,ix
      do j=1,jx

! Initialize variables
        rain(i,j) = 0.
        features = 0.
        tdel(i,j,:) = 0.
        qdel(i,j,:) = 0.
        outputs = 0.

! Combine into a vector of levels (2*n_lev_rf) 
        features(1:n_lev_rf) = real(tin(i,j,krf:kx),4)
        features(n_features) = real(phalf(i,j,kx+1),4)

! Use relative humidity or specific humidity
        if (do_rh) then
         features(n_lev_rf+1:2*n_lev_rf) = real(rh(i,j,krf:kx),4)
        else
         features(n_lev_rf+1:2*n_lev_rf) = real(qin(i,j,krf:kx),4)
        endif


! Scale inputs
        if (x_scaler) then
         features = (features - xscale_mean) / xscale_stnd
        endif

! surface pressure was set to zero when training if no_ps is true
        if (no_ps) then 
         features(n_features) = 0
        endif


! calculate predicted values using RF

! loop over trees in forest
        do itree=1,n_trees

! set current node to root node
         inode = 1 

! loop as long as not at leaf node
         do while (children_left(itree,inode) .ne. children_right(itree,inode))
          if (features(split_feature(itree,inode)).le.threshold(itree,inode)) then
            next_node = children_left(itree,inode)
          else
            next_node = children_right(itree,inode)
          endif
          inode = next_node
          if (inode .gt. n_nodes(itree)) call error_mesg ('rf_convection',  &
                        'inconsistent tree', FATAL)
         enddo

         outputs = outputs + values_predicted(itree, inode, :)
  
        enddo ! end loop over trees

        outputs = outputs/n_trees


! Separate out outputs into heating and moistening tendencies
        tdel(i,j,krf:kx) = real(outputs(1:n_lev_rf),8)
        qdel(i,j,krf:kx) = real(outputs(n_lev_rf + 1:2*n_lev_rf),8)


! Calculate precipitation rate
        do k=1, kx
            rain(i,j) = rain(i,j) - qdel(i,j,k)*(phalf(i,j,k+1)- &
                       phalf(i,j,k))/grav
        end do

       end do
     end do

   end subroutine rf_convection



!##############################################################################

  subroutine check(status)

    ! checks error status after each netcdf, prints out text message each time
    !   an error code is returned. 

    integer, intent(in) :: status

    if(status /= nf90_noerr) then
       write(*, *) trim(nf90_strerror(status))
    end if
  end subroutine check


end module rf_convection_mod

