import src.ml_load as ml_load
from netCDF4 import Dataset
import numpy as np
import pickle



def build_training_dataset(expt, t_step, t_beg, t_end, n_lon_samp=5):
    """Builds training and testing datasets from an
       idealized GCM run folder. Assumes tendencies are stored for
       instantaneous values once per day. 
    Args:
     expt (str): Path to the experiment folder
     t_step (int): Number of days between when each file is saved
     t_beg (int): Date of first time being saved
     t_end (int): Date of last time being saved
     n_lon_samp (int): Number of random longitude samples to take at each lat
                       at each time step. Default value is 5 (T42 resolution)
    """

    input_dir = '/net/esker/archive1/pog/fms_output/'
    output_dir = '/net/esker/archive1/pog/mldata/training_data/'

    # seed random number generation for reproducibility
    np.random.seed(123)


    # determine size of arrays
    if expt == 'abs1.0_abs1.4_norf_ras': # special case where train using combination of two experiments
     expt_temp = 'abs1.0_norf_ras'
    else:
     expt_temp = expt
    filename = input_dir + expt_temp + '/history/day' + \
         str(t_beg).zfill(4) + 'h00/day' + \
         str(t_beg).zfill(4) + 'h00.4xday.nc'
    print(filename)
    f = Dataset(filename, mode='r')
    half_lev = f.variables['bk'][:]
    lat = f.variables['lat'][:]
    lon = f.variables['lon'][:]
    n_lev = len(half_lev)-1
    n_lat = len(lat)
    n_lon = len(lon)
    f.close()

    # Initialize
    file_days = np.arange(t_beg, t_end+t_step, t_step)  
    if expt == 'abs1.0_abs1.4_norf_ras': # special case where train using combination of two experiments
     n_files = np.size(file_days)*2
     file_days = np.concatenate((file_days, file_days))
    else:
     n_files = np.size(file_days)
    Tin = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    qin = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    psin = np.zeros((t_step, n_lat, n_lon_samp, n_files))
    Tout = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    qout = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    Pout = np.zeros((t_step, n_lat, n_lon_samp, n_files))
    Tout_all = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    qout_all = np.zeros((t_step, n_lev, n_lat, n_lon_samp, n_files))
    Pout_all = np.zeros((t_step, n_lat, n_lon_samp, n_files))
    # Loop over files in experiment folder (assumes stats stored daily)
    for i, file_day in enumerate(file_days):
        print(file_day)
        # Initialize
        zTin = np.zeros((t_step, n_lev, n_lat, n_lon))
        zqin = np.zeros((t_step, n_lev, n_lat, n_lon))
        zpsin = np.zeros((t_step, n_lat, n_lon))
        zTout = np.zeros((t_step, n_lev, n_lat, n_lon))
        zqout = np.zeros((t_step, n_lev, n_lat, n_lon))
        zPout = np.zeros((t_step, n_lat, n_lon))
        zTout_all = np.zeros((t_step, n_lev, n_lat, n_lon))
        zqout_all = np.zeros((t_step, n_lev, n_lat, n_lon))
        zPout_all = np.zeros((t_step, n_lat, n_lon))
        # Get filename
        if expt == 'abs1.0_abs1.4_norf_ras':
          if i+1<=n_files/2:
            filename = input_dir + 'abs1.0_norf_ras/history/day' + \
               str(file_day).zfill(4) + 'h00/day' + \
               str(file_day).zfill(4) + 'h00.1xday.nc'
          else:
            filename = input_dir + 'abs1.4_norf_ras/history/day' + \
               str(file_day).zfill(4) + 'h00/day' + \
               str(file_day).zfill(4) + 'h00.1xday.nc'
        else:
          filename = input_dir + expt + '/history/day' + \
             str(file_day).zfill(4) + 'h00/day' + \
             str(file_day).zfill(4) + 'h00.1xday.nc'
        print(filename)
        # Open file and grab variables from it
        f = Dataset(filename, mode='r')
        # n_time x n_lev x n_lat x n_lon
        zTin = f.variables['t_intermed'][:]
        zqin = f.variables['q_intermed'][:]
        zpsin = f.variables['ps_intermed'][:]
        zTout = f.variables['dt_tg_convection'][:]
        zqout = f.variables['dt_qg_convection'][:]
        zPout = f.variables['convection_rain_inst'][:]  # n_time x n_lat x n_lon
        zTout_all = zTout + f.variables['dt_tg_condensation'][:]
        zqout_all = zqout + f.variables['dt_qg_condensation'][:]
        zPout_all = zPout + f.variables['condensation_rain_inst'][:]
        f.close()
        # n_time x n_lev x n_lat x n_lon ->
        # n_time x n_lev x n_lat x n_lon x n_file
        # Loop over time steps in a given file
        for k in range(zTin.shape[0]):
            # Loop over latitudes
            for j in range(zTin.shape[2]):
                # Randomly choose a few longitudes
                ind_lon = np.random.randint(0, zTin.shape[3], n_lon_samp)
                # Numpy has some strange behavior when indexing and slicing are
                # combined. See: http://stackoverflow.com/q/27094438
                Tin[k, :, j, :, i] = zTin[k, :, j, :][:, ind_lon]
                qin[k, :, j, :, i] = zqin[k, :, j, :][:, ind_lon]
                Tout[k, :, j, :, i] = zTout[k, :, j, :][:, ind_lon]
                qout[k, :, j, :, i] = zqout[k, :, j, :][:, ind_lon]
                Tout_all[k, :, j, :, i] = zTout_all[k, :, j, :][:, ind_lon]
                qout_all[k, :, j, :, i] = zqout_all[k, :, j, :][:, ind_lon]
                psin[k, j, :, i] = zpsin[k, j, ind_lon]
                Pout[k, j, :, i] = zPout[k, j, ind_lon]
                Pout_all[k, j, :, i] = zPout_all[k, j, ind_lon]

    # Permute arrays to be n_lev x n_lat x n_lon_samp x n_file x n_time
    permute = [1, 2, 3, 4, 0]
    Tin = np.transpose(Tin, permute)
    qin = np.transpose(qin, permute)
    Tout = np.transpose(Tout, permute)
    qout = np.transpose(qout, permute)
    Tout_all = np.transpose(Tout_all, permute)
    qout_all = np.transpose(qout_all, permute)
    permute = [1, 2, 3, 0]
    psin = np.transpose(psin, permute)
    Pout = np.transpose(Pout, permute)
    Pout_all = np.transpose(Pout_all, permute)
    # Reshape array to be n_lev x n_lat x t_step*n_samp*n_file
    Tin = np.reshape(Tin, (n_lev, n_lat, -1))
    qin = np.reshape(qin, (n_lev, n_lat, -1))
    Tout = np.reshape(Tout, (n_lev, n_lat, -1))
    qout = np.reshape(qout, (n_lev, n_lat, -1))
    Tout_all = np.reshape(Tout_all, (n_lev, n_lat, -1))
    qout_all = np.reshape(qout_all, (n_lev, n_lat, -1))
    psin = np.reshape(psin, (n_lat, -1))
    Pout = np.reshape(Pout, (n_lat, -1))
    Pout_all = np.reshape(Pout_all, (n_lat, -1))

    # Shuffle data and store it in separate training and testing files
    n_trn_exs = Tin.shape[2]
    randinds = np.random.permutation(n_trn_exs)
    i70 = int(0.7*np.size(randinds))
    randind_trn = randinds[:i70]
    randind_tst = randinds[i70:]
    # Store the data in files
    # For convection-only learning
    pickle.dump([Tin[:, :, randind_trn], qin[:, :, randind_trn],
                 psin[:, randind_trn],
                 Tout[:, :, randind_trn], qout[:, :, randind_trn],
                 Pout[:, randind_trn], lat, half_lev],
                open(output_dir + expt + '_conv_training.pkl', 'wb'))
    pickle.dump([Tin[:, :, randind_tst], qin[:, :, randind_tst],
                 psin[:, randind_tst],
                 Tout[:, :, randind_tst], qout[:, :, randind_tst],
                 Pout[:, randind_tst], lat, half_lev],
                open(output_dir + expt + '_conv_testing.pkl', 'wb'))
    # For convection + condensation learning
    pickle.dump([Tin[:, :, randind_trn], qin[:, :, randind_trn],
                 psin[:, randind_trn],
                 Tout_all[:, :, randind_trn], qout_all[:, :, randind_trn],
                 Pout_all[:, randind_trn], lat, half_lev],
                open(output_dir + expt + '_conv_cond_training.pkl', 'wb'))
    pickle.dump([Tin[:, :, randind_tst], qin[:, :, randind_tst],
                 psin[:, randind_tst],
                 Tout_all[:, :, randind_tst], qout_all[:, :, randind_tst],
                 Pout_all[:, randind_tst], lat, half_lev],
                open(output_dir + expt + '_conv_cond_testing.pkl', 'wb'))

def write_netcdf_rf(est_str, datasource, rain_only=False, no_cos=False, use_rh=False):
    # Set output filename
    output_filename = '/net/esker/archive1/pog/mldata/gcm_regressors/'+est_str+'.nc'
    # Load rf and preprocessors
    est, _, errors, x_ppi, y_ppi, x_pp, y_pp, lat, lev, dlev = \
        pickle.load(open('/net/esker/archive1/pog/mldata/regressors/' + est_str + '.pkl', 'rb'))

    # determine the maximum number of nodes and the number of features/outputs
    estimators = est.estimators_
    n_trees = len(estimators)
    n_nodes = np.zeros(n_trees, dtype=np.int32)
    for itree in range(n_trees):
     tree = estimators[itree].tree_
     n_nodes[itree] = tree.node_count
    max_n_nodes = np.amax(n_nodes)
    print("Maximum number of nodes across trees:")
    print(max_n_nodes)
    print("Average number of nodes across trees:")
    print(np.mean(n_nodes))
    n_features = estimators[0].tree_.n_features
    n_outputs = estimators[0].tree_.n_outputs

    # populate arrays that describe trees
    children_left = np.zeros((max_n_nodes, n_trees), dtype=np.int32)
    children_right = np.zeros((max_n_nodes, n_trees), dtype=np.int32)
    split_feature = np.zeros((max_n_nodes, n_trees), dtype=np.int32)
    n_node_samples = np.zeros((max_n_nodes, n_trees), dtype=np.int32)
    threshold = np.zeros((max_n_nodes, n_trees), dtype=np.float32)
    values_predicted = np.zeros((n_outputs, max_n_nodes, n_trees), dtype=np.float32)

    # note for python, slices don't include upper index!
    # inverse transform outputs here to speed up the GCM parameterization
    n_leaf_nodes = 0
    n_samples_leaf_nodes = 0
    for itree in range(n_trees):
     tree = estimators[itree].tree_
     children_left[:n_nodes[itree],itree] = tree.children_left
     children_right[:n_nodes[itree],itree] = tree.children_right
     split_feature[:n_nodes[itree],itree] = tree.feature
     threshold[:n_nodes[itree],itree] = tree.threshold
     n_node_samples[:n_nodes[itree],itree] =  tree.n_node_samples
     for inode in range(n_nodes[itree]):
      values_predicted[:,inode,itree] = ml_load.inverse_transform_data(y_ppi, y_pp, (tree.value[inode,:]).T, lev)
      if children_left[inode,itree]==children_right[inode,itree]: # leaf node
       n_leaf_nodes = n_leaf_nodes+1
       n_samples_leaf_nodes = n_samples_leaf_nodes+n_node_samples[inode,itree]
       

    print("Average number of leaf nodes across trees:")
    print(n_leaf_nodes/n_trees)

    # chance of not being included in bootstrap sample is (1-1/n)^n 
    # which is 1/e for large n
    # note each tree has only about (1-1/e)*n_trn_exs due to bagging
    # which is 63% of them
    # only seem to keep one when there are non-unique samples
    print("Average number of samples per leaf node:")
    print(n_samples_leaf_nodes/n_leaf_nodes)

    # Grab input and output normalization
    if x_ppi['name']=='StandardScaler':
     xscale_mean = x_pp.mean_
     xscale_stnd = x_pp.scale_
    elif x_ppi['name']!='NoScaler':
     raise ValueError('Incorrect scaler name') 

    # Write to file (use netcdf3 to be compatible with gcm; requires int32 rather than int64 which is fine)
    ncfile = Dataset(output_filename, 'w', format="NETCDF3_CLASSIC")
    # Write the dimensions
    ncfile.createDimension('dim_nodes', max_n_nodes)
    ncfile.createDimension('dim_trees', n_trees)
    ncfile.createDimension('dim_features', n_features)
    ncfile.createDimension('dim_outputs', n_outputs)
    # Create variable entries in the file
    nc_n_nodes = ncfile.createVariable('n_nodes', np.dtype('int32').char, ('dim_trees'))
    nc_children_left = ncfile.createVariable('children_left', np.dtype('int32').char, ('dim_nodes', 'dim_trees'))  
    nc_children_right = ncfile.createVariable('children_right', np.dtype('int32').char, ('dim_nodes', 'dim_trees'))  
    nc_split_feature = ncfile.createVariable('split_feature', np.dtype('int32').char, ('dim_nodes', 'dim_trees'))  
    nc_threshold = ncfile.createVariable('threshold', np.dtype('float32').char, ('dim_nodes', 'dim_trees'))  
    nc_values_predicted = ncfile.createVariable('values_predicted', np.dtype('float32').char, ('dim_outputs', 'dim_nodes', 'dim_trees'))  

    if x_ppi['name']=='StandardScaler':
     nc_xscale_mean = ncfile.createVariable('xscale_mean', np.dtype('float32').char, ('dim_features'))
     nc_xscale_stnd = ncfile.createVariable('xscale_stnd', np.dtype('float32').char, ('dim_features'))

    # Write variables and close file 
    nc_n_nodes[:] = n_nodes
    nc_children_left[:] = children_left
    nc_children_right[:] = children_right
    nc_split_feature[:] = split_feature
    nc_threshold[:] = threshold
    nc_values_predicted[:] = values_predicted

    if x_ppi['name']=='StandardScaler':
     nc_xscale_mean[:] = xscale_mean
     nc_xscale_stnd[:] = xscale_stnd

    # Write global file attributes
    ncfile.description = est_str
    ncfile.close()


def write_netcdf_nn(est_str, datasource, rain_only=False, no_cos=False, use_rh=False):

    # Set output filename
    output_filename = '/net/esker/archive1/pog/mldata/gcm_regressors/'+est_str+'.nc'
    # Load rf and preprocessors
    est, _, errors, x_ppi, y_ppi, x_pp, y_pp, lat, lev, dlev = \
        pickle.load(open('/net/esker/archive1/pog/mldata/regressors/' + est_str + '.pkl', 'rb'))
    # Need to transform some data for preprocessors to be able to export params
    x, y, _, _, _, _, _, _ = ml_load.LoadData(datasource,
                                              min_lev=min(lev),
                                              rain_only=rain_only,
                                              no_cos=no_cos,
                                              use_rh=use_rh)
    x_scl = ml_load.transform_data(x_ppi, x_pp, x, lev)
    _ = ml_load.transform_data(y_ppi, y_pp, y, lev)
    # Also need to use the predict method to be able to export ANN params
    _ = est.predict(x_scl)

   # Grab weights 
    w1 = est.get_parameters()[0].weights
    w2 = est.get_parameters()[1].weights
    b1 = est.get_parameters()[0].biases
    b2 = est.get_parameters()[1].biases

   # Grab input and output normalization
    if x_ppi['name']=='StandardScaler':
      xscale_mean = x_pp.mean_
      xscale_stnd = x_pp.scale_
    else:
      raise ValueError('Incorrect scaler name') 

    if y_ppi['name']=='SimpleY': 
      Nlev = len(lev)
      yscale = np.zeros(b2.shape)
      yscale[:Nlev] = 1.0/y_pp[0]
      yscale[Nlev:] = 1.0/y_pp[1]
    elif y_ppi['name']=='SigmaCubedY':
      Nlev = len(lev)
      yscale = np.zeros(b2.shape)
      yscale[:Nlev] = 1.0/y_pp[0]
      yscale[Nlev:] = 1.0/y_pp[1]*(lev**3)
    elif y_ppi['name']=='SigmaSquaredY':
      Nlev = len(lev)
      yscale = np.zeros(b2.shape)
      yscale[:Nlev] = 1.0/y_pp[0]
      yscale[Nlev:] = 1.0/y_pp[1]*(lev**2)
    elif y_ppi['name']=='StandardScaler':
      yscale_mean = y_pp.mean_
      yscale_stnd = y_pp.scale_
    else:
      raise ValueError('Incorrect scaler name') 

    # Write weights to file
    ncfile = Dataset(output_filename, 'w', format="NETCDF3_CLASSIC")
    # Write the dimensions
    ncfile.createDimension('N_in', w1.shape[0])
    ncfile.createDimension('N_h1', w1.shape[1])
    ncfile.createDimension('N_out', w2.shape[1])
    # Create variable entries in the file
    nc_w1 = ncfile.createVariable('w1', np.dtype('float32').char,
                                  ('N_h1', 'N_in'))  # Reverse dims
    nc_w2 = ncfile.createVariable('w2', np.dtype('float32').char,
                                  ('N_out', 'N_h1'))
    nc_b1 = ncfile.createVariable('b1', np.dtype('float32').char,
                                  ('N_h1'))
    nc_b2 = ncfile.createVariable('b2', np.dtype('float32').char,
                                  ('N_out'))
    nc_xscale_mean = ncfile.createVariable('xscale_mean',
                                           np.dtype('float32').char, ('N_in'))
    nc_xscale_stnd = ncfile.createVariable('xscale_stnd',
                                           np.dtype('float32').char, ('N_in'))
    if y_ppi['name']=='SimpleY' or y_ppi['name']=='SigmaCubedY':
     nc_yscale = ncfile.createVariable('yscale',
                                             np.dtype('float32').char,
                                             ('N_out'))
    else:
      nc_yscale_mean = ncfile.createVariable('yscale_mean', np.dtype('float32').char, ('N_out'))
      nc_yscale_stnd = ncfile.createVariable('yscale_stnd', np.dtype('float32').char, ('N_out'))


    # Write variables and close file - transpose because fortran reads it in
    # "backwards"
    nc_w1[:] = w1.T
    nc_w2[:] = w2.T
    nc_b1[:] = b1
    nc_b2[:] = b2
    nc_xscale_mean[:] = xscale_mean
    nc_xscale_stnd[:] = xscale_stnd

    if y_ppi['name']=='SimpleY' or y_ppi['name']=='SigmaCubedY':
      nc_yscale[:] = yscale
    else:
      nc_yscale_mean[:] = yscale_mean
      nc_yscale_stnd[:] = yscale_stnd


    # Write global file attributes
    ncfile.description = est_str
    ncfile.close()


